// Copyright (C) 1999-2002 Open Source Telecom Corporation.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.

#include <cc++/common.h>
#include "private.h"

using std::string;

#ifdef	CCXX_NAMESPACES
namespace ost {
#endif

#ifdef	HAVE_MODULES
DSO	*DSO::first = NULL;
DSO	*DSO::last = NULL;
Mutex	DSO::mutex;

void DSO::dynunload(void)
{
	while(DSO::last)
		delete DSO::last;
}

#ifdef	HAVE_DLFCN_H
extern "C" {
#include <dlfcn.h>
};

#ifndef	RTLD_GLOBAL
#define	RTLD_GLOBAL	0
#endif


DSO::~DSO()
{
	MutexLock lock(mutex);
	if(image)
		dlclose(image);

	if(first == this && last == this)
		first = last = NULL;

	if(!next && !prev)
		return;

	if(prev)
		prev->next = next;

	if(next)
		next->prev = prev;
	
	if(first == this)
		first = next;
	if(last == this)
		last = prev;
}

void DSO::loader(const char *filename, bool flag)
{
	next = prev = NULL;

	if(flag)
		image = dlopen(filename, RTLD_NOW | RTLD_GLOBAL);
	else
		image = dlopen(filename, RTLD_LAZY | RTLD_GLOBAL);

	if(!image)
	{
		if(getException() == Thread::throwObject)
			throw(this);
#ifdef	COMMON_STD_EXCEPTION
		else if(getException() == Thread::throwException)
			throw(DSOException(string(filename) + dlerror()));
#endif
		return;
	}

	if(!last)
	{
		last = first = this;
		return;
	}

	mutex.enterMutex();
	last->next = this;
	prev = last;
	last = this;
	mutex.leaveMutex();
}	


bool DSO::isValid(void)
{
	if(!image)
		return false;

	return true;
}

char *DSO::getError(void)
{
	return (char *)dlerror();
}

void *DSO::operator[](const char *sym)
{
	return dlsym(image, (char *)sym);
}

#endif
#endif

#ifdef	CCXX_NAMESPACES
};
#endif

/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
