// Copyright (C) 1999-2002 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.

#include <cc++/common.h>
#include <process.h>

#ifdef	CCXX_NAMESPACES
namespace ost {
using namespace std;
#endif

typedef	unsigned (__stdcall *exec_t)(void *);

typedef enum { 
	THREAD_TYPE_NORMAL=0, 
	THREAD_TYPE_MAIN, 
	THREAD_TYPE_POSIX,
	THREAD_TYPE_DUMMY
};

class MainThread : public Thread
{
protected:
	void run(void) {return;};
public:
	MainThread() : Thread(true) {};
};

static ThreadKey _self;
static MainThread _mainthread;

Thread *Thread::_main = NULL;

// invalid pointer to thread used to test deleted thread
// point in the middle of mainthread...
#define DUMMY_INVALID_THREAD ((Thread*)(((char*)((Thread*)&_mainthread))+1))

#if defined(_MSC_VER) && !defined(_DLL)
# error This project cannot be compiled as a static library. Some implementation stuff require DLL
#endif

class DummyThread: public Thread
{
private:
	void run(void) {return;};
public:
	DummyThread():Thread(false) { };
	static void CheckDelete();
};

void DummyThread::CheckDelete()
{
	Thread *th = (Thread*)_self.getKey();
	if (!th) return;

	// delete if dummy thread
	if (th->_type == THREAD_TYPE_DUMMY)
		delete th;
}

CCXX_EXPORT(Thread*) getThread(void)
{
	Thread *th = (Thread *)_self.getKey();
	if (th == DUMMY_INVALID_THREAD) return NULL;
	// for no common c++ thread construct a dummy thread
	if (!th)
		th = new DummyThread();
	return th;
}

CCXX_EXPORT(DWORD) waitThread(HANDLE hRef, timeout_t timeout)
{
	Thread *th = getThread();

	if(th)
		return th->waitHandle(hRef, timeout);
	else
		return WaitForSingleObject(hRef, timeout);
}


CCXX_EXPORT(void) ccxx_sleep(timeout_t timeout)
{
	Thread *th = getThread();
	if(th)
		th->sleep(timeout);
	else
		SleepEx(timeout, FALSE);
}

CCXX_EXPORT(void) ccxx_yield(void)
{
	Thread *th = getThread();
	if(th)
		th->yield();
	else
		SleepEx(0, FALSE);
}

Thread::Thread(bool isMain):
	_tid(GetCurrentThreadId()),
	_cancel(cancelDefault),
	_start(NULL),
	_msgpos(0),
	_throw(throwObject),
	_active(false),
	_suspendEnable(true),
	_type(THREAD_TYPE_DUMMY)
{
	// FIXME: error handling
	HANDLE process = GetCurrentProcess();
	DuplicateHandle(process,GetCurrentThread(),process,&_hThread,0,FALSE,DUPLICATE_SAME_ACCESS);
	_parent = this;
	_cancellation = CreateEvent(NULL, TRUE, FALSE, NULL);
	
	if(isMain)
	{
		_type   = THREAD_TYPE_MAIN;
		_main = this;
	}
	_self.setKey(this);
};

Thread::Thread(int pri, unsigned stack):
	_tid(0),
	_cancel(cancelDefault),
	_start(NULL),
	_msgpos(0),
	_throw(throwObject),
	_active(false),
	_suspendEnable(true),
	_type(THREAD_TYPE_NORMAL)
{
	if(!_main)
	{
		_self.setKey(NULL);
		_main = this;
	}

	_parent = getThread();
	if(_parent)
		_throw = _parent->_throw;
	else
		_parent = this;

	_cancellation = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(!_cancellation)
		throw(this);

	_hThread = (HANDLE)_beginthreadex(NULL, stack, (exec_t)&Execute, (void *)this, CREATE_SUSPENDED, (unsigned *)&_tid);
	if(!_hThread)
	{
		CloseHandle(_cancellation);
		_cancellation = NULL;
		throw(this);
	}

	setCancel(cancelInitial);

	switch(pri)
	{
	case 1:
		pri = THREAD_PRIORITY_ABOVE_NORMAL;
		break;
	case -1:
		pri = THREAD_PRIORITY_BELOW_NORMAL;
		break;
	case 2:
		pri = THREAD_PRIORITY_HIGHEST;
		break;
	case -2:
		pri = THREAD_PRIORITY_LOWEST;
		break;
	default:
		pri = THREAD_PRIORITY_NORMAL;
	}

	SetThreadPriority(_hThread, pri);
};

Thread::~Thread()
{
	terminate();

	if (isThread())
	{
		_tid = 0;
		_self.setKey(DUMMY_INVALID_THREAD);
	}
}

void Thread::exit()
{
	if (isThread())
	{
		setCancel(cancelDisabled);
		close();
		ExitThread(0);
	}
}

bool Thread::isRunning()
{
	return (_tid != 0 && _active) ? true : false;
}

void Thread::suspend(void)
{
	if (!_suspendEnable)
	{
		if (getException() != throwNothing)
			throw this;
		return;
	}
	SuspendThread(_hThread);
}

void Thread::resume(void)
{
	if (!_suspendEnable)
	{
		if (getException() != throwNothing)
			throw this;
		return;
	}
	ResumeThread(_hThread);
}

int Thread::start(Semaphore *start)
{
	_start = start;
	if (!_active)
	{
		_active = true;
		ResumeThread(_hThread);
	}
	return 0;
}	

void Thread::terminate(void)
{
	if(!_tid)
		return;

	if(_tid == GetCurrentThreadId())
		return;

	bool terminated = false;
	if(!_active)
	{
		// NOTE: add a test in testthread for multiple suspended Terminate
		ResumeThread(_hThread);
		TerminateThread(_hThread, 0);
		terminated = true;
	}
	else
	{
		switch(_cancel)	{
		case cancelImmediate:
			TerminateThread(_hThread, 0);
			terminated = true;
			break;
		default:
			SetEvent(_cancellation);
		}
	}
	WaitForSingleObject(_hThread, INFINITE);
	if(_parent)
		_parent->notify(this);
	// FIXME: test race condition, a thread can be restarted
	CloseHandle(_hThread);
	_hThread = NULL;
	CloseHandle(_cancellation);
	_cancellation = NULL;
	_tid = 0;
	if (terminated)
		final();
}

unsigned __stdcall Thread::Execute(Thread *th)
{
	_self.setKey(th);
	th->yield();
	if(!th->_active)
	{
		th->setCancel(cancelImmediate);
		th->yield();
		return 0;	
	}
	if(th->_start)
	{
		th->_start->wait();
		th->_start = NULL;
	}

	try
	{
		th->initial();
		if(th->getCancel() == cancelInitial) 
			th->setCancel(cancelDefault);
		th->run();
	}
	// ignore cancellation exception
	catch(const InterruptException&)
	{ ; }

	th->close();
	return 0;
}

void Thread::close()
{
	setCancel(cancelDisabled);
	if(_parent)
		_parent->notify(this);

	final();

	// test if this class have been deleted
	if (_self.getKey() == this)
	{
		_tid = 0;
	}
}

void Thread::yield(void)
{
	sleep(0);
}

bool Thread::testCancel(void)
{
	switch(_cancel)
	{
	case cancelInitial:
	case cancelDisabled:
		break;
	default:
		if(WaitForSingleObject(_cancellation, 0) == WAIT_OBJECT_0)
		{
			if (_cancel == cancelManual)
				throw InterruptException();
			else
				exit();
		}
	}
	return false;
}

void Thread::sleep(timeout_t timeout) 
{
	switch(_cancel)
	{
	case cancelInitial:
	case cancelDisabled:
		SleepEx(timeout, FALSE);
		break;
	default:
		if(WaitForSingleObject(_cancellation, timeout) == WAIT_OBJECT_0)
		{
			if (_cancel == cancelManual)
				throw InterruptException();
			else
				exit();
		}
	}
}

DWORD Thread::waitHandle(HANDLE obj, timeout_t timeout) 
{
	HANDLE objects[2];
	DWORD stat;

	objects[0] = _cancellation;
	objects[1] = obj;

	// FIXME: what should happen if someone enable cancellation on wait??
	switch(_cancel)
	{
	case cancelInitial:
	case cancelDisabled:
		return WaitForSingleObject(obj, timeout);
	default:
		switch(stat = WaitForMultipleObjects(2, objects, false, timeout)) 	{
		case WAIT_OBJECT_0:
			if (_cancel == cancelManual)
				throw InterruptException();
			else
				exit();
		case WAIT_OBJECT_0 + 1:
			return WAIT_OBJECT_0;
		default:
			return stat;
		}
	}
}

void Thread::setCancel(Cancel cancel)
{
	switch(cancel)	
	{
	case cancelDeferred:
	case cancelImmediate:
		_cancel = cancel;
		yield();
		break;
	case cancelDisabled:
	case cancelInitial:
		_cancel = cancel;
	}
}	

void	Thread::setSuspend(Suspend mode)
{
	_suspendEnable = (mode == suspendEnable);
}

// win32 do not support detach
int	Thread::detach(Semaphore *start)
{
	if(start)
		start->post();
	return 0;
}

// Entry point linked for default disable thread call, not suitable for threading library...
BOOL WINAPI DllMain(
        HANDLE  hDllHandle,
        DWORD   dwReason,
        LPVOID  lpreserved
        )
{
	switch(dwReason)
	{
	case DLL_THREAD_DETACH:
		DummyThread::CheckDelete();
		break;
	}
        return TRUE ;
}

#ifdef	CCXX_NAMESPACES
};
#endif

/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
